@@ -14,8 +14,12 @@ import (
 	"math"
 	"math/rand"
 	"os"
+	"strconv"
 	"testing"
 	"time"
+
+	"github.com/andybalholm/brotli/matchfinder"
+	"github.com/xyproto/randomstring"
 )
 
 func checkCompressedData(compressedData, wantOriginalData []byte) error {
@@ -67,46 +71,6 @@ func TestEncoderEmptyWrite(t *testing.T) {
 		t.Errorf("Close()=%v, want nil", err)
 	}
 }
-
-func TestWriter(t *testing.T) {
-	for level := BestSpeed; level <= BestCompression; level++ {
-		// Test basic encoder usage.
-		input := []byte("<html><body><H1>Hello world</H1></body></html>")
-		out := bytes.Buffer{}
-		e := NewWriterOptions(&out, WriterOptions{Quality: level})
-		in := bytes.NewReader([]byte(input))
-		n, err := io.Copy(e, in)
-		if err != nil {
-			t.Errorf("Copy Error: %v", err)
-		}
-		if int(n) != len(input) {
-			t.Errorf("Copy() n=%v, want %v", n, len(input))
-		}
-		if err := e.Close(); err != nil {
-			t.Errorf("Close Error after copied %d bytes: %v", n, err)
-		}
-		if err := checkCompressedData(out.Bytes(), input); err != nil {
-			t.Error(err)
-		}
-
-		out2 := bytes.Buffer{}
-		e.Reset(&out2)
-		n2, err := e.Write(input)
-		if err != nil {
-			t.Errorf("Write error after Reset: %v", err)
-		}
-		if n2 != len(input) {
-			t.Errorf("Write() after Reset n=%d, want %d", n2, len(input))
-		}
-		if err := e.Close(); err != nil {
-			t.Errorf("Close error after Reset (copied %d) bytes: %v", n2, err)
-		}
-		if !bytes.Equal(out.Bytes(), out2.Bytes()) {
-			t.Error("Compressed data after Reset doesn't equal first time")
-		}
-	}
-}
-
 func TestIssue22(t *testing.T) {
 	f, err := os.Open("testdata/issue22.gz")
 	if err != nil {
@@ -147,6 +111,45 @@ func TestIssue22(t *testing.T) {
 	}
 }
 
+func TestWriterV2(t *testing.T) {
+	for level := BestSpeed; level <= BestCompression; level++ {
+		// Test basic encoder usage.
+		input := []byte("<html><body><H1>Hello world</H1></body></html>")
+		out := bytes.Buffer{}
+		e := NewWriterV2(&out, level)
+		in := bytes.NewReader([]byte(input))
+		n, err := io.Copy(e, in)
+		if err != nil {
+			t.Errorf("Copy Error: %v", err)
+		}
+		if int(n) != len(input) {
+			t.Errorf("Copy() n=%v, want %v", n, len(input))
+		}
+		if err := e.Close(); err != nil {
+			t.Errorf("Close Error after copied %d bytes: %v", n, err)
+		}
+		if err := checkCompressedData(out.Bytes(), input); err != nil {
+			t.Error(err)
+		}
+
+		out2 := bytes.Buffer{}
+		e.Reset(&out2)
+		n2, err := e.Write(input)
+		if err != nil {
+			t.Errorf("Write error after Reset: %v", err)
+		}
+		if n2 != len(input) {
+			t.Errorf("Write() after Reset n=%d, want %d", n2, len(input))
+		}
+		if err := e.Close(); err != nil {
+			t.Errorf("Close error after Reset (copied %d) bytes: %v", n2, err)
+		}
+		if !bytes.Equal(out.Bytes(), out2.Bytes()) {
+			t.Error("Compressed data after Reset doesn't equal first time")
+		}
+	}
+}
+
 func TestEncoderStreams(t *testing.T) {
 	// Test that output is streamed.
 	// Adjust window size to ensure the encoder outputs at least enough bytes
@@ -574,6 +577,30 @@ func BenchmarkEncodeLevelsReset(b *testing.B) {
 	}
 }
 
+func BenchmarkEncodeLevelsResetV2(b *testing.B) {
+	opticks, err := ioutil.ReadFile("testdata/Isaac.Newton-Opticks.txt")
+	if err != nil {
+		b.Fatal(err)
+	}
+
+	for level := BestSpeed; level <= 9; level++ {
+		buf := new(bytes.Buffer)
+		w := NewWriterV2(buf, level)
+		w.Write(opticks)
+		w.Close()
+		b.Run(fmt.Sprintf("%d", level), func(b *testing.B) {
+			b.ReportAllocs()
+			b.ReportMetric(float64(len(opticks))/float64(buf.Len()), "ratio")
+			b.SetBytes(int64(len(opticks)))
+			for i := 0; i < b.N; i++ {
+				w.Reset(ioutil.Discard)
+				w.Write(opticks)
+				w.Close()
+			}
+		})
+	}
+}
+
 func BenchmarkDecodeLevels(b *testing.B) {
 	opticks, err := ioutil.ReadFile("testdata/Isaac.Newton-Opticks.txt")
 	if err != nil {
@@ -595,3 +622,229 @@ func BenchmarkDecodeLevels(b *testing.B) {
 		})
 	}
 }
+
+func test(t *testing.T, filename string, m matchfinder.MatchFinder, blockSize int) {
+	data, err := ioutil.ReadFile(filename)
+	if err != nil {
+		t.Fatal(err)
+	}
+	b := new(bytes.Buffer)
+	w := &matchfinder.Writer{
+		Dest:        b,
+		MatchFinder: m,
+		Encoder:     &Encoder{},
+		BlockSize:   blockSize,
+	}
+	w.Write(data)
+	w.Close()
+	compressed := b.Bytes()
+	sr := NewReader(bytes.NewReader(compressed))
+	decompressed, err := ioutil.ReadAll(sr)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if !bytes.Equal(decompressed, data) {
+		t.Fatal("decompressed output doesn't match")
+	}
+}
+
+func benchmark(b *testing.B, filename string, m matchfinder.MatchFinder, blockSize int) {
+	b.StopTimer()
+	b.ReportAllocs()
+	data, err := ioutil.ReadFile(filename)
+	if err != nil {
+		b.Fatal(err)
+	}
+
+	b.SetBytes(int64(len(data)))
+	buf := new(bytes.Buffer)
+	w := &matchfinder.Writer{
+		Dest:        buf,
+		MatchFinder: m,
+		Encoder:     &Encoder{},
+		BlockSize:   blockSize,
+	}
+	w.Write(data)
+	w.Close()
+	b.ReportMetric(float64(len(data))/float64(buf.Len()), "ratio")
+	b.StartTimer()
+	for i := 0; i < b.N; i++ {
+		w.Reset(ioutil.Discard)
+		w.Write(data)
+		w.Close()
+	}
+}
+
+func TestEncodeM4(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 18, DistanceBitCost: 66}, 1<<16)
+}
+
+func TestEncodeM4Chain256(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 18, DistanceBitCost: 66, ChainLength: 256}, 1<<16)
+}
+
+func BenchmarkEncodeM4(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, DistanceBitCost: 66}, 1<<16)
+}
+
+func TestEncodeM4Chain1(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 18, ChainLength: 1, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain1(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 1, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain2(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 2, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain4(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 4, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain8(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 8, HashLen: 5, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain16(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 16, HashLen: 5, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain32(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 32, HashLen: 5, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain64(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 64, HashLen: 5, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain128(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 128, HashLen: 5, DistanceBitCost: 66}, 1<<16)
+}
+
+func BenchmarkEncodeM4Chain256(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.M4{MaxDistance: 1 << 20, ChainLength: 256, HashLen: 5, DistanceBitCost: 66}, 1<<16)
+}
+
+func TestEncodePathfinder(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 18}, 1<<16)
+}
+
+func TestEncodePathfinderChain256(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 18, ChainLength: 256}, 1<<16)
+}
+
+func BenchmarkEncodePathfinder(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20}, 1<<16)
+}
+
+func TestEncodePathfinderChain1(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 18, ChainLength: 1}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain1(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 1}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain2(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 2}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain4(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 4}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain8(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 8, HashLen: 5}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain16(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 16, HashLen: 5}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain32(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 32, HashLen: 5}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain64(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 64, HashLen: 5}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain128(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 128, HashLen: 5}, 1<<16)
+}
+
+func BenchmarkEncodePathfinderChain256(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", &matchfinder.Pathfinder{MaxDistance: 1 << 20, ChainLength: 256, HashLen: 4, MinLength: 3}, 1<<16)
+}
+
+func TestEncodeM0(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", matchfinder.M0{}, 1<<16)
+}
+
+func BenchmarkEncodeM0(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", matchfinder.M0{}, 1<<16)
+}
+
+func TestEncodeM0Lazy(t *testing.T) {
+	test(t, "testdata/Isaac.Newton-Opticks.txt", matchfinder.M0{Lazy: true}, 1<<16)
+}
+
+func BenchmarkEncodeM0Lazy(b *testing.B) {
+	benchmark(b, "testdata/Isaac.Newton-Opticks.txt", matchfinder.M0{Lazy: true}, 1<<16)
+}
+
+func TestIssue51(t *testing.T) {
+	for i := 65536; i <= 65536*4; i += 65536 {
+		t.Run("compress data length: "+strconv.Itoa(i)+"bytes", func(t *testing.T) {
+			dataStr := randomstring.HumanFriendlyString(i)
+			dataBytes := []byte(dataStr)
+			buf := bytes.Buffer{}
+			w := NewWriterV2(&buf, 4)
+
+			n, err := w.Write(dataBytes)
+			if err != nil {
+				t.Fatalf("Error while compressing data: %v", err)
+			}
+			if n != len(dataBytes) {
+				t.Fatalf("Bytes written (%d) != len(databytes) (%d)", n, len(dataBytes))
+			}
+			err = w.Close()
+			if err != nil {
+				t.Fatalf("Error closing writer: %v", err)
+			}
+
+			r := NewReader(&buf)
+			dst := make([]byte, len(dataBytes)+100)
+			p := dst
+			total := 0
+			for {
+				n1, err1 := r.Read(p)
+				if err1 != nil {
+					if err1 != io.EOF {
+						t.Fatal(err1)
+					}
+					break
+				}
+				total += n1
+				p = p[n1:]
+			}
+			if !bytes.Equal(dst[:total], dataBytes) {
+				t.Fatal("Decompressed bytes don't match")
+			}
+		})
+	}
+}
+
+func TestIssue58(t *testing.T) {
+	content := []byte("---\nthis-is-not-brotli: \"it is actually yaml\"")
+	input := bytes.NewBuffer(content)
+
+	r := NewReader(input)
+
+	buf, err := io.ReadAll(r)
+	if err == nil {
+		t.Fatalf("expected error, got none and read:\n%x\n%s\n%v", buf, buf, buf)
+	}
+}
